#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>
#include "ahp.h"
#include "libusbahp.h"

#ifdef LINUX
#include <sys/shm.h>
#include <signal.h>

static int shmkey;
static int kpid,npid;
static struct shmid_ds buf;
static char *cmdaddr;

#endif

char xdevicefile[256]=XDEVFILENAME;

int pdimper[32]=
{
    00,03,06,10,13,16,19,23,
        26,29,32,35,39,42,45,48,
        52,55,58,61,65,68,71,74,
        77,81,84,87,90,94,97,100
};


char xtype[4][20]={"PLC-R","PLC-S","RF","Cmd"};

uchar devmap[17]={
    0xff,
    0x06,0x0e,0x02,0x0a,0x01,0x09,0x05,0x0d,
    0x07,0x0f,0x03,0x0b,0x00,0x08,0x04,0x0c};

uchar udevmap[16]={
    13, 5, 3,11,15, 7, 1, 9,
        14, 6, 4,12,16, 8, 2,10};

uchar uhcmap[16]={
    'M','E','C','K','O','G','A','I',
        'N','F','D','L','P','H','B','J'};

uchar urfhc[16]={
    'M','N','O','P','C','D','A','B',
        'E','F','G','H','K','L','I','J'};

char ufun[16][10]={
    "AllUoff","AllLon","On","Off","Dim","Bri","AllLoff","7?",
    "8?","9?","PreDim","PreDim","12?","StatOn","StatOff","StatReg"};

char allmap[16]={
    1,1,0,0,0,0,1,0,
    0,0,0,0,0,0,0,0};

char states[16]="OXXODBX789PPcXOR";

/*
0   All Units Off
1   All Lights On
2   On
3   Off
4   Dim (5 bytes) (eg 5A 03 02 0B 64 for dimming A hc)
5   Bri (5 bytes)
6   All Lights off
...
a   Preset Dim (lower 16)
b   Preset Dim (upper 16)
...
d   status=on
e   status=off
f   status req
*/

char logbuf[1024];

static int shmkey;
struct xinfos *xinfo=NULL;
static int xportnum=-1;
static int xpid;

int ParseCommand(char *cmdstr);
int dumpinhex(uchar *buf,int size);
void ProcessRecd(uchar *buf, int size);
time_t getnow(void);

int setclock(void);
int flipnib(int bits);

#ifndef SIMPLEX10
//===========================================================================
int InitX10(void)
{
    char name[100]="AHP-1";

    xportnum=-1;

#ifdef LINUX
    shmkey=shmget(918582,sizeof(struct xinfos)+100,IPC_CREAT);
    if(shmkey==-1)
    {
        printf("could not attach to X10 shared memory!\n");
        return xportnum;;
    }
    xinfo=(struct xinfos *)shmat(shmkey,NULL,0);
    if(shmctl(shmkey,IPC_RMID,&buf)!=0) printf("shmctl error\n");
#else
    xinfo=(struct xinfos *)malloc(sizeof(struct xinfos)+100);
#endif

    xportnum=AHPAcquire(name);
#ifdef LINUX
    xpid=fork();
    if(xpid==0)
    {
        ReadX10();
        exit(0);
    }
#endif
    //printf("*********************************XPORTNUM=%d\n",xportnum);
    return xportnum;
}


void CloseX10(void)
{
    if(xportnum>=0)
    {
        //printf("********************release XPORTNUM=%d\n",xportnum);
        AHPRelease(xportnum);
#ifdef LINUX
        if(xpid>0) kill(xpid,SIGTERM);
        shmdt(xinfo);
#else
        free(xinfo);
#endif
    }
}


void ReadX10(void)
{
    // child process to read the AHP
    ushort size;
    int ret=1;
    uchar readbuf[101];

#ifdef LINUX
    xinfo=(struct xinfos *)shmat(shmkey,NULL,0);
#endif
    while(ret==1)
    {
        size=100;
        ret=AHPRead(xportnum, readbuf, &size);
        if(size<=100 && ret)
        {
            dumpinhex(readbuf,size);        //for debugging
            ProcessRecd(readbuf,size);
        }
    }
#ifdef LINUX
    if(ret!=1) printf("AHP returned %d\n",ret);fflush(stdout);
    shmdt(xinfo);
#endif
}


void X10deviceAction(char *buf)
{
    char cmdbuffer[31];

    //printf("rest of cmdline=(%s)\n",buf);

    if(xportnum<0)
    {
        Logit("No X10 controller on this system","");
        return;
    }
    
    if(strlen(buf)==0)
    {
        printf("Enter X10 command: HD Fun [percent for dim-bri] (eg A1 on)\n");
        fgets(cmdbuffer,20,stdin);
        StripTail(cmdbuffer);
        if(strlen(cmdbuffer)==0) return;
        
        ParseCommand(cmdbuffer);
    }
    else ParseCommand(buf);
}


//================================================================================

int ParseCommand(char *cmdstr)
{
    char hc;
    int dev;
    int fun;
    int i;
    int percent;
    char rawstr[1024];

    StripTail(cmdstr);
    strcpy(rawstr,cmdstr);
    strupr(cmdstr);

    hc=cmdstr[0];
    if(hc<'A' || hc>'P')
    {
        Logit("HC error",cmdstr);
        return 0;
    }
    i=1;
    dev=-1;
    while(cmdstr[i]==' ') i++;
    if(cmdstr[i]<='9')
    {
        dev=atoi(&cmdstr[i]);
        while(cmdstr[i]!='\0' && cmdstr[i]!=' ') i++;
        while(cmdstr[i]==' ') i++;
        if(dev<1 || dev>16)
        {
            Logit("Device Code error",cmdstr);
            return(0);
        }
    }
    fun=-1;
    if(strstr(&cmdstr[i],"ALLUOFF")==&cmdstr[i]) fun=ALLUOFF;
    else if(strstr(&cmdstr[i],"ALLLON")==&cmdstr[i]) fun=ALLLON;
    else if(strstr(&cmdstr[i],"ON")==&cmdstr[i]) fun=ON;
    else if(strstr(&cmdstr[i],"OFF")==&cmdstr[i]) fun=OFF;
    else if(strstr(&cmdstr[i],"DIM")==&cmdstr[i]) fun=DIM;
    else if(strstr(&cmdstr[i],"BRI")==&cmdstr[i]) fun=BRIGHT;
    else if(strstr(&cmdstr[i],"ALLLOFF")==&cmdstr[i]) fun=ALLLOFF;
    else if(strstr(&cmdstr[i],"STAT")==&cmdstr[i])
    {
        fun=STATREQ;
        xinfo->statdev=dev-1;
        xinfo->stathc=hc-'A';
    }
    else if(strstr(&cmdstr[i],"CLR")==&cmdstr[i]) fun=XCLR;
    else if(strstr(&cmdstr[i],"NAME")==&cmdstr[i]) fun=XNAME;
    else
    {
        Logit("Function Code error",cmdstr);
        return(0);
    }
    if(fun!=XNAME)
    {
        while(cmdstr[i]>' ') i++;
        while(cmdstr[i]==' ') i++;
        percent=atoi(&cmdstr[i]);
    }

    if(dev<0 && (fun==XCLR || fun==XNAME || allmap[fun]==0))
    {
        Logit("This function requires a Device code",cmdstr);
        return(0);
    }

    if(fun==XCLR)
    {
        xinfo->xdev[hc-'A'][dev-1].status=UNKNOWN;
        xinfo->xdev[hc-'A'][dev-1].lasttime=0;
        SaveX10Dev();
    }
    else if(fun==XNAME)
    {
        //printf("%s\n",&cmdstr[i]);
        i=i+4;
        while(cmdstr[i]==' ') i++;
        //skip escape chare from cgi for ? name (clear name)
        if(cmdstr[i]=='\\') i++;
        if(strlen(&cmdstr[i])==0)
        {
            Logit("No value for name!",cmdstr);
        }
        else
        {
            // use rawstr so that it's not all upper case!
            strcpy(xinfo->xdev[hc-'A'][dev-1].name,&rawstr[i]);
            SaveX10Dev();
        }
    }
    else
    {
        SendCommand(hc,dev,fun, percent,L_PLC_S);
    }
    return(1);
}

void TestCommand(void)
{
    /*BB FF FF 01 10 14 20 15 line
to the CM15A. I noticed that I would get a reply of 00. So I attempted
to write it again (after a short wait) and I would then get a 55.
So it looks like the CM15A sends back a NOT_OK an OK. Now things
get more interesting, I attempted to send the d1 1f e0 */
    /*
    uchar sendbuff[20];
    ushort size;
    
    size=0;
    sendbuff[size++]=0xbb;
    sendbuff[size++]=0xff;
    sendbuff[size++]=0x7f;
    sendbuff[size++]=0x01;
    sendbuff[size++]=0x10;
    sendbuff[size++]=0x14;
    sendbuff[size++]=0x20;
    sendbuff[size++]=0x15;
    sendbuff[size++]=0x20;
    sendbuff[size++]=0x24;
    AHPWrite(xportnum,sendbuff,&size);
    
    msDelay(2000);

    size=0;
    sendbuff[size++]=0x8b;
    AHPWrite(xportnum,sendbuff,&size);
    msDelay(500);
    
    size=0;
    sendbuff[size++]=0xdb;
    sendbuff[size++]=0x1f;
    sendbuff[size++]=0xe0;
    AHPWrite(xportnum,sendbuff,&size);*/

}


void SendCommand(char hc, int dev, int function,int percent,int logtype)
{
    int ihc,i;
    uchar sendbuff[10];
    ushort size=2;
    int doit;
    if(xportnum>=0)
    {
        if(hc>'Z') hc=hc-('a'-'A');
        ihc=hc-'A'+1;

        if(function==UNKNOWN)
        {
            xinfo->xdev[ihc-1][dev-1].status=function;
            SaveX10Dev();
            return;
        }

        if(!allmap[function])
        {
            sendbuff[0]=0x04;
            sendbuff[1]=(devmap[ihc]<<4)+devmap[dev];
            //printf("%02X %02X\n",sendbuff[0],sendbuff[1]);
            size=2;
            AHPWrite(xportnum,sendbuff,&size);
            msDelay(750);
        }
        sendbuff[0]=0x06;
        sendbuff[1]=(devmap[ihc]<<4)+function;
        //printf("%02X %02X\n",sendbuff[0],sendbuff[1]);
        if(function==DIM || function==BRIGHT)
        {   
            // not sure when lowest bit used or for what
            sendbuff[2]=(percent<<1)+1;
            size=3;
        }
        else size=2;
        AHPWrite(xportnum,sendbuff,&size);
        msDelay(750);

        if(function==STATREQ)
        {
            xinfo->stathc=ihc-1;
            xinfo->statdev=dev-1;
        }

        if(allmap[function])
        {
            for(i=0;i<16;i++)
            {
                doit=0;
                if(xinfo->xdev[ihc-1][i].status!=UNKNOWN) doit=1;
                if((xinfo->xdev[ihc-1][i].status==OFF || xinfo->xdev[ihc-1][i].status==ALLUOFF || xinfo->xdev[ihc-1][i].status==ALLLOFF) &&
                    (function==ALLUOFF || function==ALLLOFF)) doit=0;
                else if((xinfo->xdev[ihc-1][i].status==ON || xinfo->xdev[ihc-1][i].status==ALLLON) && function==ALLLON) doit=0;
                        
                if(doit)
                {
                    xinfo->xdev[ihc-1][i].status=function;
                    xinfo->xdev[ihc-1][i].lasttime=getnow();
                }
            }
            dev=0;
        }
        else
        {
            if(function!=STATREQ)
            {
                xinfo->xdev[ihc-1][dev-1].status=function;
                xinfo->xdev[ihc-1][dev-1].lasttime=getnow();
            }
        }
        LogX10(__LINE__,logtype,ihc-1,dev-1,function,percent);
    }
}

void SPrintX10Devices(void)
{
    int i,j;

    if(xportnum>=0)
    {
        printf("\n ***** X10 Module Status *****\n");
        for(i=0;i<16;i++)
        {
            printf("%c: ",i+'A');
            for(j=0;j<16;j++)
            {
                if(xinfo->xdev[i][j].status==UNKNOWN) printf(". ");
                else printf("%c ",states[xinfo->xdev[i][j].status]);
            }
            printf("\n");
        }
        printf("\n");
    }
}
#endif


void SaveX10Dev(void)
{
    FILE *devinfo;
    int i,j;

    devinfo=fopen(xdevicefile,"w");
    if(devinfo==NULL) return;

    for(i=0;i<16;i++)
    {
        for(j=0;j<16;j++)
        {
            fprintf(devinfo,"%c%02d %02d %012d %s\n",
                i+'A',j+1,
                xinfo->xdev[i][j].status,
                xinfo->xdev[i][j].lasttime,
                xinfo->xdev[i][j].name);
        }
    }
    fclose(devinfo);
}

void LoadX10Dev(void)
{
    char inbuf[300];
    FILE *devinfo;
    int i,j;

    if(xinfo==NULL) return;

    for(i=0;i<16;i++)
    {
        for(j=0;j<16;j++)
        {
            xinfo->xdev[i][j].status=UNKNOWN;
            xinfo->xdev[i][j].lasttime=0;
            strcpy(xinfo->xdev[i][j].name,"?");
            xinfo->xdev[i][j].hc=i;
            xinfo->xdev[i][j].dev=j;
        }
    }
    xinfo->stathc=-1;
    xinfo->statdev=-1;
    for(i=0;i<16;i++) xinfo->lastdevs[i]=-1;

    devinfo=fopen(xdevicefile,"r");
    if(devinfo==NULL) return;
    i=0;
    j=0;

    fgets(inbuf,299,devinfo);
    while(!feof(devinfo))
    {
        StripTail(inbuf);
        if(strlen(inbuf)<15) goto done;
        xinfo->xdev[i][j].status=atoi(&inbuf[4]);
        xinfo->xdev[i][j].lasttime=atoi(&inbuf[7]);
        strcpy(xinfo->xdev[i][j].name,&inbuf[20]);
        //if(strcmp(xinfo->xdev[i][j].name,"No Name")==0) strcpy(xinfo->xdev[i][j].name,"?");
        j++;
        if(j>=16)
        {
            j=0;
            i++;
        }
        if(i>16) goto done;
        fgets(inbuf,299,devinfo);
    }
    Logit("X10 Devices Loaded","");

done:
    fclose(devinfo);
}


struct xdevinfo *FindX10Device(char *name)
{
    int hc,dev;
    int found=0;
    int i,j;

    if(strcmp(name,"?")==0) return NULL;
    if(xinfo==NULL) return NULL;

    if(strlen(name)>1 && strlen(name)<4)        //hc dev specified?
    {
        hc=name[0]-'A';
        dev=atoi(&name[1])-1;
        if(hc>=0 && hc<16 && dev>=0 && dev<16)
        {
            return &xinfo->xdev[hc][dev];
        }
    }
    else                                        //look for name
    {
        for(i=0;i<16;i++)
        {
            for(j=0;j<16;j++)
            {
                if(stricmp(xinfo->xdev[i][j].name,name)==0)
                {
                    return &xinfo->xdev[i][j];
                }
            }
        }
    }
    return NULL;
}

uchar SimpleX10State(struct xdevinfo *xdev)
{
    switch(xdev->status)
    {
    case ON:
    case ALLLON:
    case STATON:
        return ON;
        break;
    case OFF:
    case ALLUOFF:
    case ALLLOFF:
    case STATOFF:
        return OFF;
        break;
    case DIM:
    case PDIMH:
    case PDIML:
    case BRIGHT:
        return DIM;
        break;
    default:
        return UNKNOWN;
        break;
    }
}



#ifndef SIMPLEX10

int dumpinhex(uchar *buf,int size)
{
#ifndef XXYYZ
    int i;

    if(buf[0]!=0x5a && buf[0]!=0x5d && buf[0]!=0x55)
    {
        printf("%s ------ Raw X10 data:",TS());
        for(i=0;i<size;i++)
        {
            printf(" %02X",buf[i]);
        }
        printf("\n");
        fflush(stdout);
    }
#endif
    return 0;
}

void ProcessRecd(uchar *buf, int size)
{
    int i;
    static int lasthc=-1;
    int hc;
    int fun;
    int doit;
    int dev;

    if(buf[0]==0xa5)
    {
        setclock();
    }

    else if(buf[0]==0x5a)       // x10 PLC data
    {
        if(buf[2]==0x00 && size==4) //device
        {
            lasthc=uhcmap[buf[3]>>4]-'A';
            xinfo->lastdevs[lasthc]=udevmap[buf[3]&0x0f]-1;
        }
        else                //command
        {
            if(size==4)
            {
                hc=uhcmap[buf[3]>>4]-'A';
                fun=buf[3]&0x0f;

                if(allmap[fun])
                {
                    LogX10(__LINE__,L_PLC_R,hc,-1,fun,0);
                    for(i=0;i<16;i++)
                    {
                        doit=0;
                        if(xinfo->xdev[hc][i].status!=UNKNOWN) doit=1;
                        if((xinfo->xdev[hc][i].status==OFF || xinfo->xdev[hc][i].status==ALLUOFF || xinfo->xdev[hc][i].status==ALLLOFF) &&
                            (fun==ALLUOFF || fun==ALLLOFF)) doit=0;
                        else if((xinfo->xdev[hc][i].status==ON || xinfo->xdev[hc][i].status==ALLLON) && fun==ALLLON) doit=0;
                        if(doit)
                        {
                            xinfo->xdev[hc][i].status=fun;
                            xinfo->xdev[hc][i].lasttime=getnow();
                        }
                    }
                    SaveX10Dev();
                }
                else
                {
                    if(fun==PDIMH || fun==PDIML)
                    {
                        int level;
                        level=flipnib(buf[3]>>4);
                        if(fun==PDIMH) level=level|=0x10;
                        //no hc contained in this..  use last
                        if(xinfo->stathc>=0 && xinfo->statdev>=0)
                        {
                            //printf("preset dimval=%d (%d) %d=%d\n",buf[3]>>4,fun==PDIM?1:0, level, pdimper[level]);

                            xinfo->xdev[xinfo->stathc][xinfo->statdev].status=fun;
                            xinfo->xdev[xinfo->stathc][xinfo->statdev].lasttime=getnow();
                            LogX10(__LINE__,L_PLC_R,xinfo->stathc,xinfo->statdev,fun,pdimper[level]);
                        }
                    }
                    else if(fun==STATON || fun==STATOFF)
                    {
                        if(xinfo->statdev>=0)
                        {
                            //printf("statdev=%d\n",xinfo->statdev);
                            xinfo->xdev[hc][xinfo->statdev].status=fun;
                            xinfo->xdev[hc][xinfo->statdev].lasttime=getnow();
                            LogX10(__LINE__,L_PLC_R,hc,xinfo->statdev,fun,0);
                        }
                    }
                    else
                    {
                        if(xinfo->lastdevs[hc]>=0)
                        {
                            if(fun==STATREQ)
                            {
                                xinfo->statdev=xinfo->lastdevs[hc];
                                xinfo->stathc=hc;
                            }
                            else
                            {
                                xinfo->xdev[hc][xinfo->lastdevs[hc]].status=fun;
                                xinfo->xdev[hc][xinfo->lastdevs[hc]].lasttime=getnow();
                            }
                            LogX10(__LINE__,L_PLC_R,hc,xinfo->lastdevs[hc],fun,0);
                        }
                        else printf("fallthru fun =%02x\n",fun);
                    }
                }
            }
            else if(size==5)        //dim and bright
            {
//printf("dim/bri\n");
                hc=uhcmap[buf[4]>>4]-'A';
                fun=buf[4]&0x0f;
                if(xinfo->lastdevs[hc]>=0)
                {
                    xinfo->xdev[hc][xinfo->lastdevs[hc]].status=fun;
                    xinfo->xdev[hc][xinfo->lastdevs[hc]].lasttime=getnow();
                    LogX10(__LINE__,L_PLC_R,hc,xinfo->lastdevs[hc],fun,-1);
                }
            }
            else
            {
                printf("Not handled! ");
                printf("size=%d\n");
                for(i=0;i<size;i++)
                {
                    printf("%02X ",buf[i]);
                }
                printf("\n");
                fflush(stdout);         
            }
        }
    }
///  under construction!

    else if(buf[0]==0x5d)
    {
#ifdef XXYYZ    
        /*
        printf("%s RF command:===========================",TS());
        for(i=0;i<size;i++)
        {
            printf("%02X ",buf[i]);
        }
        printf("\n");
        fflush(stdout);
        
*/
        hc=buf[2]>>4;
        if(!(buf[4]&0x80))
        {
            dev=0;
            if(buf[4]&0x40)dev=dev+4;
            if(buf[4]&0x10)dev=dev+1;
            if(buf[4]&0x08)dev=dev+2;
            if(buf[2]&0x0f)dev=dev+8;
            dev++;
            //printf("%s RF:  %c-%-2d %s\n",TS(),urfhc[hc],dev,buf[4]&0x20?"Off":"On");

            fun=2+((buf[4]&0x20)>0);

            xinfo->xdev[urfhc[hc]-'A'][dev-1].status=fun;
            xinfo->xdev[urfhc[hc]-'A'][dev-1].lasttime=getnow();
            LogX10(__LINE__,L_RF,urfhc[hc]-'A',dev-1,fun,X10FLAG);
        }
        else
        {
            printf("%s RF:  %c-%s\n",TS(),urfhc[hc],buf[4]&0x10?"Dim":"Bri");

            fun=4+((buf[4]&0x10)==0);
            //status[urfhc[hc]-'A'][lastdev]=states[fun];
        }
        fflush(stdout);
#endif
    }
/*
    else if(buf[0]==0x5b)
    {
        printf("%s Macro action--------------------------: ",TS());
        for(i=0;i<size;i++)
        {
            printf("%02X ",buf[i]);
        }
        printf("\n");
        fflush(stdout);
    }
    //else if(buf[0]==0x55)
    //{
        //seems to be an "end" single byte command
    //}
    else
    {
        printf("%s ???: ",TS());
        for(i=0;i<size;i++)
        {
            printf("%02X ",buf[i]);
        }
        printf("\n");
        fflush(stdout);
    }
    */
}

void LogX10(int line, int type, int hc,int dev,int fun,int percent)
{
    if(hc<0 || dev<0)
    {
        //printf("called line =%d logX10 hc=%d dev=%d\n",line,hc,dev);
    }
    
    if(allmap[fun] || fun==STATON || fun==STATOFF)
    {
        sprintf(logbuf,"%-3s: %c    %s",xtype[type],hc+'A',ufun[fun&0x0f]);
        Logit(logbuf,"");
    }
    else if((fun==DIM || fun==PDIMH || fun==PDIML) && percent>=0)
    {
        sprintf(logbuf,"%-3s: %c-%-2d %s %d%c",
            xtype[type],hc+'A',dev+1,ufun[fun&0x0f],percent,'%');
        Logit(logbuf,xinfo->xdev[hc][dev].name);
    }
    else
    {
        sprintf(logbuf,"%-3s: %c-%-2d %s",xtype[type],hc+'A',dev+1,ufun[fun&0x0f]);
        Logit(logbuf,xinfo->xdev[hc][dev].name);
    }
    SaveX10Dev();
}

time_t getnow(void)
{
    time_t now;

    time(&now);
    return now;
}


#ifdef DOC_ONLY
struct cm11_setclock {
   uchar cmd;       /* 0x9b */
   uchar seconds;
   uchar minutes;   /* 0-119          */
   uchar hours;     /* 0-11 (hours/2) */
   uchar yearday;   /* really 9 bits  */
   uchar daymask;   /* really 7 bits  */
   uchar house;     /* 0:timer purge, 1:monitor clear, 3:battery clear */
   uchar dummy;     /* Do we really need this ? */
};
#endif

int setclock(void)
{
   struct tm *tm;
   //int rtn, i;
   ushort size;

   time_t t;

   //char str[256];

   //struct cm11_setclock  sc;

   uchar sendbuff[20];

   Logit("Setting CM15a Clock","");

   size=0;
   time(&t);
   tm = localtime(&t);

   sendbuff[size++]= 0x9b;              //function code
   sendbuff[size++]= tm->tm_sec;        //seconds
   sendbuff[size++]= tm->tm_min + 60 * (tm->tm_hour & 1);       //0-199
   sendbuff[size++]= tm->tm_hour >> 1;  //0-11 (hours/2)
   sendbuff[size++]= tm->tm_yday;       //really 9 bits
   sendbuff[size]= 1 << tm->tm_wday;    //daymask (7 bits)
   if(tm->tm_yday & 0x100) //
     sendbuff[size] |= 0x80;
   size++;
   sendbuff[size++]= 0x60;              // house
   sendbuff[size++]= 0x00;          // Filler

   // -> Write time & something
   // Get nothing back!

   AHPWrite(xportnum, sendbuff, &size);
   return 1;
}


int flipnib(int bits)
{
    int ret=0;

    if(bits&0x08) ret|=0x01;
    if(bits&0x04) ret|=0x02;
    if(bits&0x02) ret|=0x04;
    if(bits&0x01) ret|=0x08;
    return ret;
}

#endif


